package com.winhc.bigdata.spark.ng.change

import com.winhc.bigdata.spark.config.EsConfig
import com.winhc.bigdata.spark.utils.BaseUtil.isWindows
import com.winhc.bigdata.spark.utils._
import org.apache.spark.internal.Logging
import org.apache.spark.sql.functions.col
import org.apache.spark.sql.types.{MapType, StringType, StructField, StructType}
import org.apache.spark.sql.{DataFrame, Row, SparkSession}

import scala.annotation.meta.getter
import scala.collection.mutable

/**
 * @Description: 筛选出数据的具体变更
 */
object NgChangeExtract {

  //判断两个map在指定key上是否相等，如不等反回不相等字段
  def getDoubleDataMap(iterable: Iterable[Map[String, String]]): (Map[String, String], Map[String, String]) = {
    val map = iterable.map(m => (m("change_flag"), m)).toMap
    (map.getOrElse("0", null), map.getOrElse("1", null))
  }

  case class ChangeExtractHandle(s: SparkSession,
                                 project: String, //表所在工程名
                                 tableName1: String, //表名（不加前后辍）
                                 primaryKey: String, //此维度主键
                                 inc_ds: String, //需要计算的分区
                                 primaryFields: Seq[String] //主要字段，该字段任意一个不同 则认为发生变化
                                ) extends LoggingUtils with Logging {
    @(transient@getter) val spark: SparkSession = s

    val target_eci_change_extract = "ads_change_extract"

    val updateTimeMapping = Map(
      "wenshu_detail_combine" -> "update_date", //文书排序时间
      "company_equity_info_list" -> "reg_date" //文书排序时间
    )
    //不同name映射table
    val tabMapping =
      Map("company_holder_v2" -> "company_holder" //胜诉案件
      )

    //转换字段
    def trans(s: String): String = {
      var res = s
      if (tabMapping.contains(s)) {
        res = tabMapping(s)
      }
      res
    }

    def calc(): Unit = {
      val tableName = trans(tableName1)
      val cols = primaryFields.filter(!_.equals(primaryKey)).seq

      val ds = inc_ds.replace("-", "")

      val intersectCols = getColumns(s"$project.ads_$tableName").toSet & getColumns(s"$project.inc_ads_$tableName").toSet

      val otherAllCols = intersectCols.filter(!primaryKey.equals(_)).toSeq
      val all_cols = primaryKey +: otherAllCols :+ "change_flag"

      val lastDs_ads_all = getLastPartitionsOrElse(s"$project.ads_$tableName", "0")

      val handle = ReflectUtils.getClazz[NgCompanyChangeHandle](s"com.winhc.bigdata.spark.ng.change.table.$tableName1", cols)

      val update_time = BaseUtil.nowDate()

      val df = sql(
        s"""
           |SELECT  $primaryKey,${otherAllCols.mkString(",")},'0' as change_flag
           |FROM    $project.inc_ads_$tableName
           |WHERE   ds = $ds
           |UNION ALL
           |SELECT  t2.$primaryKey,${otherAllCols.map("t2." + _).mkString(",")},'1' as change_flag
           |FROM    (
           |            SELECT  DISTINCT ${primaryKey}
           |            FROM    $project.inc_ads_$tableName
           |            WHERE   ds = $ds
           |        ) AS t1
           |JOIN    (
           |             SELECT  tmp.*
           |             FROM    (
           |                         SELECT  a.*
           |                                 ,row_number() OVER (PARTITION BY a.${primaryKey} ORDER BY ds DESC, ${updateTimeMapping.getOrElse(tableName, "update_time")} DESC) c
           |                         FROM    (
           |                                     SELECT  ${intersectCols.mkString(",")},ds
           |                                     FROM    $project.ads_$tableName
           |                                     WHERE   ds = $lastDs_ads_all
           |                                     UNION ALL
           |                                     SELECT  ${intersectCols.mkString(",")},ds
           |                                     FROM    $project.inc_ads_$tableName
           |                                     WHERE   ds > $lastDs_ads_all and ds < $ds
           |                                 ) AS a
           |                     ) AS tmp
           |             WHERE   tmp.c = 1
           |        ) AS t2
           |ON      t1.${primaryKey} = t2.${primaryKey}
           |""".stripMargin)


      val rdd =
        df.select(all_cols.map(column => col(column).cast("string")): _*)
          .rdd.map(r => {
          (r.getAs[String](primaryKey), all_cols.map(f => (f, r.getAs[String](f))).toMap)
        }).groupByKey()
          .map(x => {
            val rowkey = x._1
            val map_list = x._2
            val m = getDoubleDataMap(map_list)

            val new_map = m._1
            val old_map = m._2
            if (new_map == null && old_map == null) {
              null
            } else if (old_map == null) {
              val res = handle.handle(rowkey, null, map_list.head)
              if (res == null) {
                null
              } else {
                Row(res._1, res._2, tableName, res._3, res._4, res._5, res._6, res._7, res._8, update_time, res._9)
              }
            } else if (new_map == null) {
              null
            } else {
              val res = handle.handle(rowkey, old_map, new_map)
              if (res == null) {
                null
              } else {
                Row(res._1, res._2, tableName, res._3, res._4, res._5, res._6, res._7, res._8, update_time, res._9)
              }
            }
          }).filter(_ != null)

      val schema = StructType(Array(
        StructField("rowkey", StringType), //表数据主建
        StructField("company_id", StringType), //公司id
        StructField("table_name", StringType), //表名
        StructField("type", StringType), // 变更类型 insert update
        StructField("data", MapType(StringType, StringType)), //变更后数据
        StructField("fields", StringType), //如果是更新 则显示更新字段
        StructField("title", StringType), // 动态数据展示 ps. 新增某土地公示
        StructField("label", StringType), // 1.一般变更，2.风险变更
        StructField("biz_time", StringType), //业务时间
        StructField("update_time", StringType), //处理时间
        StructField("old_data", MapType(StringType, StringType)) //变更前数据
      ))

      spark.createDataFrame(rdd, schema)
        .createOrReplaceTempView(s"tmp_change_extract_view_$tableName1")

      sql(
        s"""
           |INSERT ${if (isWindows) "INTO" else "OVERWRITE"} TABLE ${project}.$target_eci_change_extract PARTITION(ds='$ds',tn='$tableName1')
           |SELECT *
           |FROM
           |    tmp_change_extract_view_$tableName1
           |""".stripMargin)
    }
  }


  private val startArgs = Seq(
    Args(tableName = "company_holder", primaryFields = "percent,deleted")
    , Args(tableName = "company_staff", primaryFields = "staff_type,deleted")
    , Args(tableName = "company", primaryKey ="company_id", primaryFields = "name,cate_third_code,county_code,reg_capital_amount")
  )


  private case class Args(project: String = "winhc_ng"
                          , tableName: String
                          , primaryKey: String = "rowkey"
                          , primaryFields: String)


  def main(args: Array[String]): Unit = {
    val Array(tableName, inc_ds) = args

    val config = EsConfig.getEsConfigMap ++ mutable.Map(
      "spark.hadoop.odps.project.name" -> "winhc_ng",
      "spark.hadoop.odps.spark.local.partition.amt" -> "100"
    )
    val spark = SparkUtils.InitEnv("NgChangeExtract", config)

    var start = startArgs
    if (!tableName.equals("all")) {
      val set = tableName.split(",").toSet
      start = start.filter(a => set.contains(a.tableName))
    }

    val a = start.map(e => (e.tableName, () => {
      ChangeExtractHandle(spark, e.project, e.tableName, e.primaryKey, inc_ds, e.primaryFields.split(",")).calc()
      true
    }))

    AsyncExtract.startAndWait(spark, a)

    spark.stop()
  }

}
